from collections import namedtuple, OrderedDict
import copy
import logging
import json
import os
import re

import boto3
from botocore.exceptions import ClientError
from jinja2 import Environment, DictLoader, FileSystemLoader
import jsonref
from lambda_uploader import package, uploader
from retrying import retry
import ruamel.yaml as yaml

from .build_deps import PythonDependencyBuilder
from . import templates
from . import utils

LOG = logging.getLogger(__name__)

API_GATEWAY_URL_TEMPLATE = "https://{}.execute-api.{}.amazonaws.com/{}"


class Deployment(object):

    def __init__(self, config):
        self.config = config
        self.project_dir = self.config['project_dir']
        self.stage = config['stage']
        self.region = self.config['stages'][self.stage]['region']
        self.lambda_path = os.path.join(self.project_dir,
                                        self.config['Lambda']['path'])
        self.account_id = config['account_id']
        self.extra_files = self.normalize_extra_files(config['Lambda'])
        # Let's make sure the accounts match up
        self.verify_account_id()

    def apply_templates(self, template):
        aws_int = 'x-amazon-apigateway-integration'
        paths = template.get('paths')
        for path, methods in paths.items():
            for method, _config in methods.items():
                if _config.get('x-yoke-integration'):
                    template['paths'][path][method][
                        aws_int] = self.template_aws_integration(
                            _config['x-yoke-integration'])
        return template

    def build_lambda_package(self, skip_if_exists=False):
        LOG.warning("Building Lambda package ...")
        pkg = package.Package(self.lambda_path)
        pkg._requirements_file = None
        if os.path.isfile(pkg.zip_file):
            if skip_if_exists:
                # Package already built, don't do anything else
                LOG.warning(
                    "Lambda package already built, using existing file.")
                return pkg
            else:
                # Otherwise we should delete the existing file, otherwise it
                # will be packaged up.
                LOG.warning("Removing existing Lambda package.")
                os.remove(pkg.zip_file)

        dependency_config = self.config['Lambda'].get('dependencies')
        if dependency_config is not None:
            build_enabled = dependency_config.get('build', False)
            if build_enabled:
                LOG.warning("Building dependencies enabled ...")
                runtime = self.config['Lambda']['config'].get(
                    'runtime', 'python2.7')
                if not runtime.startswith('python'):
                    raise Exception(
                        "Building dependencies only supported on Python "
                        "runtimes."
                    )
                service_name = self.config['Lambda']['config']['name']
                wheelhouse_path = dependency_config.get('wheelhouse')
                if wheelhouse_path is not None:
                    wheelhouse_path = os.path.abspath(
                        os.path.join(
                            wheelhouse_path,
                            service_name,
                        ),
                    )
                else:
                    wheelhouse_path = os.path.abspath(
                        os.path.join(
                            self.project_dir,
                            '../../wheelhouse',
                            service_name,
                        ),
                    )
                install_dir = dependency_config.get('install_dir') or './lib'
                builder = PythonDependencyBuilder(
                    runtime=runtime,
                    project_path=self.project_dir,
                    wheelhouse_path=wheelhouse_path,
                    lambda_path=self.lambda_path,
                    install_dir=install_dir,
                    service_name=service_name,
                    extra_packages=dependency_config.get('packages'),
                )
                builder.build()

        if self.extra_files:
            for _file in self.extra_files:
                pkg.extra_file(_file)
        pkg._prepare_workspace()
        pkg.package()
        pkg.clean_workspace()
        return pkg

    def create_upldr_config(self):
        Lambda = self.config['Lambda']
        Lambda['config']['s3_bucket'] = None
        Lambda['config']['publish'] = True
        Lambda['config']['alias'] = self.stage
        Lambda['config']['alias_description'] = Lambda['config']['description']
        Lambda['config']['region'] = self.region
        Lambda['config']['runtime'] = Lambda['config'].get('runtime',
                                                           'python2.7')
        Lambda['config']['variables'] = {}
        Lambda['config']['raw'] = {'vpc': Lambda['config'].get('vpc')}
        if not Lambda['config'].get('tracing'):
            Lambda['config']['tracing'] = {}
        ordered = OrderedDict(sorted(Lambda['config'].items(),
                                     key=lambda x: str(x[1])))
        upldr_config = namedtuple('config', ordered.keys())(**ordered)
        return upldr_config

    def deploy_lambda(self):
        self.write_lambda_json()
        self.write_lambda_config()
        pkg = self.build_lambda_package(skip_if_exists=True)

        # Create fake config for lambda uploader because
        # it isn't completely ok with being used as a library.
        upldr_config = self.create_upldr_config()

        # Upload lambda
        self.upload_lambda(pkg, upldr_config)

    @retry(retry_on_exception=utils.retry_if_api_limit,
           wait_exponential_multiplier=5000, wait_exponential_max=25000,
           stop_max_attempt_number=10)
    def deploy_api(self):

        # Template swagger.yml from template
        template = self.render_swagger()
        self.write_template(template)

        # Import/Update API from swagger.yml
        # Convert to JSON and deref for AWS API compatibility.
        upload_body = self.deref(template)
        api = self.upload_api(upload_body)
        LOG.warning("Deploying API to %s stage ...", self.stage)
        client = boto3.client('apigateway', region_name=self.region)
        deployment = client.create_deployment(
            restApiId=api['id'],
            stageName=self.stage)
        LOG.warning("API deployed! URL:\n %s",
                    API_GATEWAY_URL_TEMPLATE.format(api['id'],
                                                    self.region,
                                                    self.stage))
        return deployment

    def normalize_extra_files(self, lambda_config):
        normalized = []
        extra_files = lambda_config.get('extraFiles', [])
        for extra_file in extra_files:
            normalized.append(os.path.join(self.project_dir, extra_file))
        return normalized

    def render_swagger(self):
        LOG.warning("Templating swagger.yml for region %s ...", self.region)
        swagger_file = self.config['apiGateway'].get('swaggerTemplate',
                                                     'template.yml')
        j2_env = Environment(loader=FileSystemLoader(self.project_dir),
                             trim_blocks=True, lstrip_blocks=True)
        first_template = yaml.safe_load(
            j2_env.get_template(swagger_file).render(
                accountId=self.account_id,
                Lambda=self.config['Lambda'],
                apiGateway=self.config['apiGateway'],
                region=self.region,
                stage=self.stage))

        integrations_template = self.apply_templates(first_template)

        # We have to do this twice to template the integrations - I'm sorry.
        j2_env = Environment(loader=DictLoader(
            {'template': json.dumps(integrations_template)}))
        j2_template = j2_env.get_template('template')
        rendered_template = yaml.safe_load(j2_template.render(
            accountId=self.account_id,
            Lambda=self.config['Lambda'],
            apiGateway=self.config['apiGateway'],
            region=self.region,
            stage=self.stage
        ))

        return rendered_template

    def template_aws_integration(self, yoke_integration):
        integ = copy.deepcopy(templates.AWS_INTEGRATION)
        location = 'requestTemplates'
        content_type = 'application/json'
        integ[location] = copy.deepcopy(templates.DEFAULT_REQUESTS)
        integ[location][content_type] = self.template_operation(
            integ['requestTemplates']['application/json'],
            yoke_integration.get('operation'),
        )
        integ['responses'] = copy.deepcopy(templates.DEFAULT_RESPONSES)
        return integ

    def template_operation(self, template, operation):
        p = re.compile(".*?\{\{ (.*?) \}\}.*?")
        match = p.findall(template)
        for var in match:
            template = template.replace("{{{{ {} }}}}".format(var),
                                        operation, 1)
        return template

    def upload_api(self, upload_body):
        LOG.warning("Uploading API to AWS Account %s for region %s ...",
                    self.account_id, self.region)
        client = boto3.client('apigateway', region_name=self.region)

        # Try to find API by name
        apis = client.get_rest_apis()
        apis = apis.get('items')
        api = None
        for item in apis:
            if self.config['apiGateway']['name'] == item['name']:
                LOG.warning('Found existing API: %s', item['name'])
                api = item
                break

        parameters = {'basepath': 'prepend'}
        if api:
            LOG.warning("API %s already exists - updating ...", api['name'])
            api = client.put_rest_api(restApiId=api['id'],
                                      mode='overwrite',
                                      body=json.dumps(upload_body),
                                      parameters=parameters)
        else:
            LOG.warning("API %s not found, importing ...",
                        self.config['apiGateway']['name'])
            api = client.import_rest_api(body=json.dumps(upload_body),
                                         parameters=parameters)

        return api

    def upload_lambda(self, pkg, upldr_config):
        LOG.warning("Uploading Lambda %s to AWS Account %s "
                    "for region %s ...",
                    upldr_config.name, self.account_id, upldr_config.region)
        upldr = uploader.PackageUploader(upldr_config, None)
        upldr.upload(pkg)
        upldr.alias()
        pkg.clean_zipfile()

    def verify_account_id(self):
        LOG.warning('Verifying AWS Account Credentials ...')

        try:
            aws_account_id = boto3.client('iam').get_user()[
                'User']['Arn'].split(':')[4]
        except ClientError:
            aws_account_id = boto3.client('sts').get_caller_identity()[
                'Account']
        try:
            assert aws_account_id == self.account_id
        except Exception:
            LOG.error('yoke.yml accountId (%s) does not match credentials'
                      'account (%s)!', self.account_id, aws_account_id)
            raise

    def write_lambda_config(self):
        lambda_config = self.config['stages'][self.stage].get('config')
        if lambda_config:
            config_file = os.path.join(self.lambda_path,
                                       'config.json')
            with open(config_file, 'w') as outfile:
                json.dump(lambda_config, outfile)

    def write_lambda_json(self):
        lambda_json = os.path.join(self.lambda_path,
                                   'lambda.json')
        with open(lambda_json, 'w') as outfile:
            json.dump(self.config['Lambda']['config'], outfile)

    def write_template(self, output, filename=None):
        if not filename:
            filename = 'swagger.yml'
        swagger_file = os.path.join(self.project_dir, filename)
        _, ext = os.path.splitext(filename)
        with open(swagger_file, 'w') as fh:
            # Could be `.yaml` or `.yml` :/
            if '.y' in ext:
                fh.write(yaml.round_trip_dump(output))
            elif '.json' in ext:
                fh.write(json.dumps(output))
        return swagger_file

    def deref(self, data):
        """AWS doesn't quite have Swagger 2.0 validation right and will fail
        on some refs. So, we need to convert to deref before
        upload."""

        # We have to make a deepcopy here to create a proper JSON
        # compatible object, otherwise `json.dumps` fails when it
        # hits jsonref.JsonRef objects.
        deref = copy.deepcopy(jsonref.JsonRef.replace_refs(data))

        # Write out JSON version because we might want this.
        self.write_template(deref, filename='swagger.json')

        return deref



import argparse
import logging
import os
import sys

from . import config
from . import deploy
from . import utils

LOG = logging.getLogger(__name__)


def build(args):
    LOG.warning('Building deployment only ...')
    deployment = deploy.Deployment(args.config)
    deployment.write_lambda_json()
    deployment.write_lambda_config()
    if args.config.get('apiGateway'):
        template = deployment.render_swagger()
        swagger_file = deployment.write_template(template)
        # Also write the deref'd JSON version
        deployment.deref(template)
        LOG.warning('API Gateway Swagger file written to {}'.format(
                    swagger_file))

    deployment.build_lambda_package()


def decrypt(args):
    utils.decrypt(args.config, output=True)


def deploy_app(args):
    deployment = deploy.Deployment(args.config)
    deployment.deploy_lambda()
    if args.config.get('apiGateway'):
        deployment.deploy_api()
    LOG.warning('Deployment complete!')


def encrypt(args):
    utils.encrypt(args.config, output=True)


def main(arv=None):
    parser = argparse.ArgumentParser(
        description='AWS Lambda + API Gateway Deployment Tool'
    )

    subparsers = parser.add_subparsers()

    parser.add_argument('--debug', dest='loglevel', help='Debug logging',
                        action='store_const', const=logging.DEBUG)
    parser.set_defaults(loglevel=logging.WARNING)

    deploy_parser = subparsers.add_parser(
        'deploy',
        help='Deploy lambda and (optionally) API Gateway.')
    deploy_parser.add_argument('--stage', dest='stage', help='Stage to deploy',
                               default=os.getenv('YOKE_STAGE'))
    deploy_parser.add_argument('--environment', '-e', dest='environment',
                               help=('Extra config values for lambda '
                                     'config - can be used multiple times'),
                               default=[], action='append',
                               metavar='KEYNAME=VALUE')
    deploy_parser.add_argument('project_dir', default=os.getcwd(), nargs='?',
                               help='Project directory containing yoke.yml')
    deploy_parser.set_defaults(func=deploy_app)

    build_parser = subparsers.add_parser(
        'build',
        help='Only template config files and build lambda package.')
    build_parser.add_argument('--stage', dest='stage', help='Stage to build',
                              default=os.getenv('YOKE_STAGE'))
    build_parser.add_argument('--environment', '-e', dest='environment',
                              help=('Extra config values for lambda '
                                    'config - can be used multiple times'),
                              default=[], action='append',
                              metavar='KEYNAME=VALUE')
    build_parser.add_argument('project_dir', default=os.getcwd(), nargs='?',
                              help='Project directory containing yoke.yml')
    build_parser.set_defaults(func=build)

    decrypt_parser = subparsers.add_parser('decrypt')
    decrypt_parser.add_argument('--stage', dest='stage',
                                help='Stage to decrypt',
                                default=os.getenv('YOKE_STAGE'))
    decrypt_parser.add_argument('project_dir', default=os.getcwd(), nargs='?',
                                help='Project directory containing yoke.yml')
    decrypt_parser.set_defaults(func=decrypt)

    encrypt_parser = subparsers.add_parser('encrypt')
    encrypt_parser.add_argument('--stage', dest='stage',
                                help='Stage to encrypt',
                                default=os.getenv('YOKE_STAGE'))
    encrypt_parser.add_argument('project_dir', default=os.getcwd(), nargs='?',
                                help='Project directory containing yoke.yml')
    encrypt_parser.set_defaults(func=encrypt)

    args = parser.parse_args()

    if args.loglevel is logging.WARNING:
        format = "%(message)s"
        logging.basicConfig(level=args.loglevel, format=format,
                            stream=sys.stdout)
    else:
        logging.basicConfig(level=args.loglevel, stream=sys.stdout)

    try:
        args.project_dir = os.path.abspath(args.project_dir)
        if hasattr(args, 'environment'):
            env_dict = utils.format_env(args.environment)
        else:
            env_dict = {}
        _cfg = config.YokeConfig(args, args.project_dir, args.stage, env_dict)
        args.config = _cfg.get_config()
        args.func(args)
    except Exception:
        LOG.exception('ERROR!')
        sys.exit(1)



import re


class HttpVerb:
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    PATCH = 'PATCH'
    HEAD = 'HEAD'
    DELETE = 'DELETE'
    OPTIONS = 'OPTIONS'
    ALL = '*'


class AuthPolicy(object):

    def __init__(self, token, aws_account_id, principal=None):
        self.token = token
        # AWS account ID policy will be generated for
        self.aws_account_id = aws_account_id
        # Principal used for policy, unique ID for end user
        self.principal_id = principal or ''
        # Policy version should always be '2012-10-17'
        self.version = '2012-10-17'
        # Used to validate resource paths for policy
        self.path_regex = '^[/.a-zA-Z0-9-\*]+$'
        # Lists of allowed/denied methods, objects with resource ARN and
        # nullable conditions statement
        self.allowMethods = []
        self.denyMethods = []
        # The API Gateway API id. By default this is set to '*'
        self.rest_api_id = '*'
        # The region where the API is deployed. Default is '*'
        self.region = '*'
        # The name of the stage used in the policy. Default is '*'
        self.stage = '*'

    def _add_method(self, effect, verb, resource, conditions):
        """
        Adds a method to the internal lists of allowed or denied methods.
        Each object in the internal list contains a resource ARN and a
        condition statement. The condition statement can be null.
        """
        if verb != '*' and not hasattr(HttpVerb, verb):
            raise NameError('Invalid HTTP verb ' + verb +
                            '. Allowed verbs in HttpVerb class')
        resource_pattern = re.compile(self.path_regex)
        if not resource_pattern.match(resource):
            raise NameError('Invalid resource path: ' + resource +
                            '. Path should match ' + self.path_regex)

        if resource[:1] == '/':
            resource = resource[1:]

        resource_arn = ('arn:aws:execute-api:' +
                        self.region + ':' +
                        self.aws_account_id + ':' +
                        self.rest_api_id + '/' +
                        self.stage + '/' +
                        verb + '/' +
                        resource)

        if effect.lower() == 'allow':
            self.allowMethods.append({
                'resource_arn': resource_arn,
                'conditions': conditions
            })
        elif effect.lower() == 'deny':
            self.denyMethods.append({
                'resource_arn': resource_arn,
                'conditions': conditions
            })

    def _get_empty_statement(self, effect):
        """
        Returns an empty statement object prepopulated with the
        correct action and the desired effect.
        """
        statement = {
            'Action': 'execute-api:Invoke',
            'Effect': effect[:1].upper() + effect[1:].lower(),
            'Resource': []
        }

        return statement

    def _get_effect_statement(self, effect, methods):
        """
        This function loops over an array of objects containing
        a resourceArn and conditions statement and generates
        the array of statements for the policy.
        """
        statements = []

        if len(methods) > 0:
            statement = self._get_empty_statement(effect)

            for method in methods:
                if (method['conditions'] is None or
                        len(method['conditions']) == 0):
                    statement['Resource'].append(method['resource_arn'])
                else:
                    cond_statement = self._get_empty_statement(effect)
                    cond_statement['Resource'].append(method['resource_arn'])
                    cond_statement['Condition'] = method['conditions']
                    statements.append(cond_statement)
            statements.append(statement)

        return statements

    def allow_all_methods(self):
        """Adds a '*' allow to authorize access to all methods of an API"""
        self._add_method('Allow', HttpVerb.ALL, '*', [])

    def deny_all_methods(self):
        """Adds a '*' allow to deny access to all methods of an API"""
        self._add_method('Deny', HttpVerb.ALL, '*', [])

    def allow_method(self, verb, resource):
        """
        Adds an API Gateway method (Http verb + Resource path)
        to the list of allowed methods for the policy
        """
        self._add_method('Allow', verb, resource, [])

    def deny_method(self, verb, resource):
        """
        Adds an API Gateway method (Http verb + Resource path)
        to the list of denied methods for the policy
        """
        self._add_method('Deny', verb, resource, [])

    def allow_method_with_conditions(self, verb, resource, conditions):
        """
        Adds an API Gateway method (Http verb + Resource path) to the
        list of allowed methods and includes a condition for the policy
        statement. More on AWS policy conditions here:
        http://docs.aws.amazon.com/IAM/latest/UserGuide/
        reference_policies_elements.html#Condition
        """
        self._add_method('Allow', verb, resource, conditions)

    def deny_method_with_conditions(self, verb, resource, conditions):
        """
        Adds an API Gateway method (Http verb + Resource path) to the
        list of denied methods and includes a condition for the policy
        statement. More on AWS policy conditions here:
        http://docs.aws.amazon.com/IAM/latest/UserGuide/
        reference_policies_elements.html#Condition
        """
        self._add_method('Deny', verb, resource, conditions)

    def build(self):
        """
        Generates the policy document based on the internal lists of
        allowed and denied conditions. This will generate a policy with
        two main statements for the effect: one statement for Allow and
        one statement for Deny. Methods that includes conditions will
        have their own statement in the policy.
        """
        if ((self.allowMethods is None or len(self.allowMethods) == 0) and
                (self.denyMethods is None or len(self.denyMethods) == 0)):
            raise NameError('No statements defined for the policy')

        policy = {
            'principalId': self.principal_id,
            'policyDocument': {
                'Version': self.version,
                'Statement': []
            }
        }

        policy['policyDocument']['Statement'].extend(
            self._get_effect_statement('Allow', self.allowMethods))
        policy['policyDocument']['Statement'].extend(
            self._get_effect_statement('Deny', self.denyMethods))

        return policy



from hashlib import sha1
import logging
import os
import subprocess
from tempfile import mkstemp
import time

import docker

from .templates import DOCKER_BUILD_SCRIPT
from .templates import DOCKER_INSTALL_SCRIPT

LOG = logging.getLogger(__name__)

BUILD_IMAGE = "quay.io/pypa/manylinux1_x86_64"
CONTAINER_POLL_INTERVAL = 10
PYTHON_VERSION_MAP = {
    # cp27-mu is compiled with ucs4 support which is the same as Lambda
    'python2.7': 'cp27-cp27mu',
    'python3.6': 'cp36-cp36m',
}
STATUS_EXITED = 'exited'


def wait_for_container_to_finish(container):
    while container.status != STATUS_EXITED:
        time.sleep(CONTAINER_POLL_INTERVAL)
        container.reload()

    exit_code = container.attrs['State']['ExitCode']
    if exit_code != 0:
        # Save logs for further inspection -- if we are on CircleCI, save the
        # file under the artifacts directory.
        basepath = (
            os.environ['CIRCLE_ARTIFACTS']
            if 'CIRCLECI' in os.environ
            else '.'
        )
        log_filename = os.path.join(
            basepath,
            'container_{}.log'.format(container.short_id),
        )
        log_contents = container.logs(stdout=True, stderr=True)
        with open(log_filename, 'w') as fp:
            try:
                fp.write(log_contents)
            except TypeError:
                # On Python 3, `fp.write()` expects a string instead of bytes
                # (which is coming out of the `logs()` call), but Python 2
                # can't handle writing unicode to a file, so we can't do this
                # in both cases.
                fp.write(log_contents.decode('utf-8'))

        raise Exception(
            "Container exited with non-zero code. Logs saved to {}".format(
                log_filename)
        )


def remove_container(container):
    try:
        container.remove()
    except:
        # We just log an error and swallow the exception, because this happens
        # often on CircleCI.
        LOG.error(
            "Could not remove container, please remove it manually (ID: %s)",
            container.short_id,
        )


class PythonDependencyBuilder(object):

    def __init__(self, runtime, project_path, wheelhouse_path, lambda_path,
                 install_dir, service_name, extra_packages=None):
        """Initialize dependency builder object.

        :param runtime: Lambda Python runtime version.
        :param project_path: Full path to the project root, where `yoke.yml`
        exists.
        :param wheelhouse_path: Full path to the directory that will hold
        the build depdendencies (wheels).
        :param lambda_path: Full path to the directory where the Lambda
        function's code lives.
        :param install_dir: Relative path within the project root where the
        dependencies will be installed.
        :param service_name: Name of the service, used for creating some
        sub-directories and files.
        :param extra_packages: List of extra CentOS packages that should be
        installed before the dependencies are built.
        """
        self.runtime = runtime
        self.project_path = project_path
        self.wheelhouse_path = wheelhouse_path
        self.lambda_path = lambda_path
        self.install_dir = install_dir
        self.service_name = service_name
        self.extra_packages = extra_packages or []

    def should_rebuild(self):
        # There's a way to force rebuilding of dependencies
        if os.environ.get('FORCE_WHEEL_REBUILD') == 'true':
            LOG.warning(
                "FORCE_WHEEL_REBUILD env variable set, rebuilding "
                "dependencies.")
            return True
        # There's also a way to remove all existing dependencies, to force a
        # rebuild that way.
        sha1sum_file = os.path.join(self.wheelhouse_path, 'sha1sum')
        if os.environ.get('FORCE_WHEEL_CLEANUP') == 'true':
            LOG.warning(
                "FORCE_WHEEL_CLEANUP env variable set, cleaning up dependency "
                "directory.")
            wheel_pattern = os.path.join(self.wheelhouse_path, '*.whl')
            subprocess.call('rm -f {}'.format(wheel_pattern), shell=True)
            subprocess.call(['rm', '-f', sha1sum_file])

        if not os.path.isfile(sha1sum_file):
            LOG.warning(
                "SHA1 checksum file does not exist, rebuilding dependencies.")
            return True
        with open(sha1sum_file, 'r') as fp:
            sha1sum = fp.read().strip()

        requirements_file = os.path.join(self.lambda_path, 'requirements.txt')
        with open(requirements_file, 'r') as fp:
            calculated_sha1sum = sha1(fp.read().encode('utf-8')).hexdigest()

        if sha1sum != calculated_sha1sum:
            LOG.warning("SHA1 mismatch, rebuilding dependencies.")
            return True
        else:
            LOG.warning("SHA1 match, skip building dependencies.")
            return False

    def build(self):
        if not self.should_rebuild():
            return

        try:
            # Allow connecting to older Docker versions (e.g. CircleCI 1.0)
            client = docker.from_env(version='auto')
        except:
            LOG.error("Docker is not running, or it's outdated.")
            raise

        # Build dependencies
        build_script_path = self.generate_build_script()
        try:
            container = client.containers.run(
                image=BUILD_IMAGE,
                command='/bin/bash -c "./build_wheels.sh"',
                detach=True,
                environment={
                    'EXTRA_PACKAGES': ' '.join(self.extra_packages),
                    'PY_VERSION': PYTHON_VERSION_MAP[self.runtime],
                },
                volumes={
                    self.wheelhouse_path: {'bind': '/wheelhouse'},
                    self.lambda_path: {'bind': '/src'},
                    build_script_path: {'bind': '/build_wheels.sh'},
                },
            )
            LOG.warning(
                "Build container started, waiting for completion (ID: %s)",
                container.short_id,
            )
            wait_for_container_to_finish(container)
            LOG.warning("Build finished.")
            remove_container(container)
        finally:
            os.remove(build_script_path)

        # Install dependencies
        install_script_path = self.generate_install_script()
        try:
            project_path = os.path.dirname(self.lambda_path)
            container = client.containers.run(
                image=BUILD_IMAGE,
                command='/bin/bash -c "./install_wheels.sh"',
                detach=True,
                environment={
                    'INSTALL_DIR': self.install_dir,
                    'PY_VERSION': PYTHON_VERSION_MAP[self.runtime],
                },
                volumes={
                    self.wheelhouse_path: {'bind': '/wheelhouse'},
                    project_path: {'bind': '/src'},
                    install_script_path: {'bind': '/install_wheels.sh'},
                },
            )
            LOG.warning(
                "Install container started, waiting for completion (ID: %s)",
                container.short_id,
            )
            wait_for_container_to_finish(container)
            LOG.warning("Install finished.")
            remove_container(container)
        finally:
            os.remove(install_script_path)

    def _generate_script(self, contents):
        # Place temporary file inside `/tmp`, because Docker for Mac only
        # allows mounting volumes from certain locations, and the default temp
        # directory is not among them.
        fd, filename = mkstemp(suffix='.sh', dir='/tmp')
        os.close(fd)
        with open(filename, 'w') as fp:
            fp.write(contents)

        os.chmod(filename, 0o755)
        return filename

    def generate_build_script(self):
        return self._generate_script(DOCKER_BUILD_SCRIPT)

    def generate_install_script(self):
        return self._generate_script(DOCKER_INSTALL_SCRIPT)



import json


def load_config():
    with open('config.json', 'r') as config_file:
        raw = config_file.read()
    return json.loads(raw)


def lambda_handler(event, context):
    config = load_config()
    print(event)
    return {"message": config['message']}



import boto3
from botocore.exceptions import ClientError
import json
import logging
import os
import re
import ruamel.yaml as yaml

from . import utils

LOG = logging.getLogger(__name__)
LAMBDA_ROLE_ARN_TEMPLATE = "arn:aws:iam::{account_id}:role/{role}"


class YokeConfig(object):

    def __init__(self, shellargs, project_dir, stage, env_dict):
        self._args = shellargs
        self.project_dir = project_dir
        self.stage = stage
        self.env_dict = env_dict
        self.yoke_path = os.path.join(self.project_dir, 'yoke.yml')

    def check_default_stage(self, config, stage):
        # Set provided stage's config to default configs
        if stage == 'default':
            config['stages'][self.stage] = config['stages'][stage]
            # If you don't have a default config, initialize it.
            if not config['stages'][self.stage].get('config'):
                config['stages'][stage]['config'] = {}
            stage = self.stage
        config['stage'] = self.stage
        return config

    def get_config(self):
        config = self.load_config_file()
        stage = self.get_stage(self.stage, config)
        config = self.check_default_stage(config, stage)

        config['project_dir'] = self.project_dir
        config['account_id'] = self.get_account_id()

        if 'config' not in config['stages'][self.stage]:
            config['stages'][self.stage]['config'] = {}

        if self._args.func.__name__ not in ['encrypt', 'decrypt']:
            if (config['stages'][self.stage].get('secret_config') or
                    config['stages'][self.stage].get('secretConfig')):
                dec_config = utils.decrypt(config)
                config['stages'][self.stage]['config'].update(dec_config)

        if self.env_dict:
            config['stages'][self.stage]['config'].update(self.env_dict)

        # Set proper Lambda ARN for role
        config['Lambda']['config']['role'] = LAMBDA_ROLE_ARN_TEMPLATE.format(
            account_id=config['account_id'],
            role=config['Lambda']['config']['role']
        )

        LOG.info('Config:\n%s', json.dumps(config, indent=4))

        return config

    def get_stage(self, stage, config):
        assert stage, "No YOKE_STAGE envvar found - aborting!"
        LOG.warning("Loading stage %s ...", stage)
        if not config['stages'].get(stage):
            if not config['stages'].get('default'):
                LOG.warning('%s stage was not found and no default '
                            'stage was provided - aborting!', stage)
                raise
            LOG.warning("%s stage was not found - using default ...", stage)
            stage = 'default'
        return stage

    def get_account_id(self):
        LOG.warning('Getting AWS Account Credentials ...')
        try:
            aws_account_id = boto3.client('iam').get_user()[
                'User']['Arn'].split(':')[4]
        except ClientError as exc:
            LOG.debug("Failed to get account via get_user()...\n %s",
                      str(exc))
            aws_account_id = boto3.client('sts').get_caller_identity()[
                'Account']

        return str(aws_account_id)

    def load_config_file(self):
        # Read config file lines as list in order to template.
        LOG.warning("Getting config from %s ...", self.project_dir)
        with open(self.yoke_path, 'r') as config_file:
            raw = config_file.readlines()
        raw = self.render_config(raw)
        return yaml.safe_load(raw)

    def render_config(self, config):
        vars = self.env_dict
        vars['stage'] = self.stage
        rendered = []
        p = re.compile(".*?\{\{ (.*?) \}\}.*?")
        for line in config:
            match = p.findall(line)
            for var in match:
                replace_var = self.env_dict[var]
                line = line.replace("{{{{ {} }}}}".format(var),
                                    replace_var, 1)
            rendered.append(line)
        LOG.debug("Rendered config:\n{}".format(rendered))
        return ''.join(rendered)



import base64
import json
import logging

import boto3
import ruamel.yaml as yaml

LOG = logging.getLogger(__name__)


def check_encryption_required_fields(stage):
    for field in ['keyRegion', 'keyName']:
        if field not in stage:
            raise Exception("`{}` is a required field when secretConfig is "
                            "used.".format(field))


def decrypt(config, output=False):
    stage = config['stage']
    check_encryption_required_fields(config['stages'][stage])

    enc_config = get_secret_config(config, stage)
    if not isinstance(enc_config, basestring):
        raise Exception('Secret config for stage {} is not a '
                        'string! Did you forget to encrypt '
                        'first?'.format(stage))
    stage_cfg = base64.b64decode(enc_config)
    region = config['stages'][stage]['keyRegion']
    kms = boto3.client('kms', region_name=region)
    resp = kms.decrypt(CiphertextBlob=bytes(stage_cfg))
    plain = json.loads(resp['Plaintext'])
    if output:
        print('Decrypted config for stage {}:\n\n{}'.format(
            stage,
            yaml.round_trip_dump(plain)))
    return plain


def encrypt(config, output=False):
    stage = config['stages'][config['stage']]
    check_encryption_required_fields(stage)
    secret_config = get_secret_config(config, config['stage'])
    kms = boto3.client('kms', region_name=stage['keyRegion'])
    key_name = 'alias/{}'.format(stage['keyName'])
    resp = kms.encrypt(KeyId=key_name,
                       Plaintext=bytes(json.dumps(secret_config)))
    if output:
        print('Encrypted config for stage {}:\nsecretConfig: "{}"\n'.format(
            config['stage'],
            base64.b64encode(resp['CiphertextBlob'])))


def format_env(env_list):
    env_dict = {}
    for env_item in env_list:
        # A value might contain an '=' so let's not clobber that.
        parts = env_item.split('=')
        key = parts.pop(0)
        value = '='.join(parts)
        env_dict[key] = value
    return env_dict


def get_secret_config(config, stage):
    old_style = config['stages'][stage].get('secret_config')
    if old_style:
        LOG.warning('{} stage is using "secret_config" - please update to'
                    ' "secretConfig". This will break in the '
                    'future!'.format(stage))
    new_style = config['stages'][stage].get('secretConfig')
    if old_style and new_style:
        raise Exception('Please use only one of '
                        '"secretConfig" or "secret_config!"')
    return old_style if old_style else new_style


def retry_if_api_limit(exception):
    if 'TooManyRequestsException' in str(exception):
        LOG.warning('Hit API Gateway rate limit - retrying ...')
        return True
    return False



import json
import sys
import traceback

import policy


def load_config():
    with open('config.json', 'r') as config_file:
        raw = config_file.read()
    return json.loads(raw)


def lambda_handler(event, context):  # pylint: disable=unused-argument
    try:
        tmp = event['methodArn'].split(':')
        api_gateway_arn = tmp[5].split('/')
        account_id = tmp[4]

        config = load_config()

        expected_token = config['expected_token']
        token = event['authorizationToken']

        authpolicy = policy.AuthPolicy(token, account_id)
        authpolicy.rest_api_id = api_gateway_arn[0]
        authpolicy.region = tmp[3]
        authpolicy.stage = api_gateway_arn[1]

        if token == expected_token:
            authpolicy.allow_all_methods()
        else:
            authpolicy.deny_all_methods()

        return authpolicy.build()
    except Exception:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback.print_exception(exc_type, exc_value, exc_traceback)
        raise Exception("Unauthorized")



import json

from fleece.httperror import HTTPError


def load_config():
    with open('config.json', 'r') as config_file:
        raw = config_file.read()
    return json.loads(raw)


def lambda_handler(event, context):
    try:
        config = load_config()
        print(event)
        return {"message": config['message']}
    except Exception as exc:
        print(str(exc))
        raise HTTPError(status=500, message=str(exc))



import ast
import re
from setuptools import setup, find_packages


INSTALL_REQUIRES = [
    'boto3>=1.4.2',
    'botocore>=1.4.85',
    'docker>=2.0.0',
    'Jinja2>=2.8',
    'jsonref>=0.1',
    'lambda-uploader>=1.1.0',
    'retrying>=1.3.3',
    'ruamel.yaml>=0.11.11'
]


def package_meta():
    """Read __init__.py for global package metadata.
    Do this without importing the package.
    """
    _version_re = re.compile(r'__version__\s+=\s+(.*)')
    _url_re = re.compile(r'__url__\s+=\s+(.*)')
    _license_re = re.compile(r'__license__\s+=\s+(.*)')

    with open('yoke/__init__.py', 'rb') as ffinit:
        initcontent = ffinit.read()
        version = str(ast.literal_eval(_version_re.search(
            initcontent.decode('utf-8')).group(1)))
        url = str(ast.literal_eval(_url_re.search(
            initcontent.decode('utf-8')).group(1)))
        license = str(ast.literal_eval(_license_re.search(
            initcontent.decode('utf-8')).group(1)))
    return {
        'version': version,
        'license': license,
        'url': url,
    }


meta = package_meta()


setup(
    name='yoke',
    description='AWS Lambda + API Gateway Deployer',
    keywords='aws lambda apigateway',
    version=meta['version'],
    packages=find_packages(exclude=['tests', 'examples']),
    test_suite='tests',
    install_requires=INSTALL_REQUIRES,
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
    ],
    license=license,
    author="Rackers",
    maintainer_email="ryan.walker@rackspace.com",
    url=meta['url'],
    entry_points={
        'console_scripts': [
            'yoke=yoke.shell:main'
        ]
    },
)



# flake8: noqa

APPLICATION_JSON_RESPONSE_FMT = (
    '{"error": {"code": %(rc)s, "message": $input.json(\'$.errorMessage\')}}'
)
RESPONSE_CODES = [
    300,
    301,
    302,
    303,
    304,
    305,
    307,
    400,
    401,
    402,
    403,
    404,
    405,
    406,
    407,
    408,
    409,
    410,
    411,
    412,
    413,
    414,
    415,
    416,
    417,
    418,
    422,
    423,
    500,
    501,
    502,
    503,
    504,
    505,
]
DEFAULT_RESPONSES = {
    '^{rc}:.*'.format(rc=resp_code): {
        'responseTemplates': {
            'application/json': (
                APPLICATION_JSON_RESPONSE_FMT % dict(rc=resp_code)
            ),
        },
        'statusCode': '{rc}'.format(rc=resp_code),
    }
    for resp_code in RESPONSE_CODES
}
DEFAULT_RESPONSES['default'] = {
    'responseTemplates': {
        'application/json': '__passthrough__'
    },
    'statusCode': '200',
}

APPLICATION_JSON_REQUEST = """\
{
  "rawContext": {
    "apiId": "$context.apiId",
    "authorizer": {
      "principalId": "$context.authorizer.principalId",
      "claims": {
        "property": "$context.authorizer.claims.property"
      }
    },
    "httpMethod": "$context.httpMethod",
    "identity": {
      "accountId": "$context.identity.accountId",
      "apiKey": "$context.identity.apiKey",
      "caller": "$context.identity.caller",
      "cognitoAuthenticationProvider": "$context.identity.cognitoAuthenticationProvider",
      "cognitoAuthenticationType": "$context.identity.cognitoAuthenticationType",
      "cognitoIdentityId": "$context.identity.cognitoIdentityId",
      "cognitoIdentityPoolId": "$context.identity.cognitoIdentityPoolId",
      "sourceIp": "$context.identity.sourceIp",
      "user": "$context.identity.user",
      "userAgent": "$context.identity.userAgent",
      "userArn": "$context.identity.userArn"
    },
    "requestId": "$context.requestId",
    "resourceId": "$context.resourceId",
    "resourcePath": "$context.resourcePath",
    "stage": "$context.stage"
  },
  "operation": "{{ operation }}",
  "parameters": {
    "gateway": {
      "id": "$context.apiId",
      "stage": "$context.stage",
      "request-id" : "$context.requestId",
      "resource-path" : "$context.resourcePath",
      "http-method": "$context.httpMethod",
      "stage-data": {
        #foreach($param in $stageVariables.keySet())
        "$param": "$util.escapeJavaScript($stageVariables.get($param))" #if($foreach.hasNext),#end
        #end
      }
    },
    "requestor": {
      "source-ip": "$context.identity.sourceIp",
      "user-agent": "$context.identity.userAgent",
      "account-id" : "$context.identity.accountId",
      "api-key" : "$context.identity.apiKey",
      "caller": "$context.identity.caller",
      "user": "$context.identity.user",
      "user-arn" : "$context.identity.userArn"
    },
    "request": {
      "querystring": {
        #foreach($param in $input.params().querystring.keySet())
        "$param": "$util.escapeJavaScript($input.params().querystring.get($param))"#if($foreach.hasNext),#end
        #end
      },
      "path": {
        #foreach($param in $input.params().path.keySet())
        "$param": "$util.escapeJavaScript($input.params().path.get($param))" #if($foreach.hasNext),#end
        #end
      },
      "header": {
        #foreach($param in $input.params().header.keySet())
        "$param": "$util.escapeJavaScript($input.params().header.get($param))" #if($foreach.hasNext),#end
        #end
      },
      "body": $input.json('$')
    }
  }
}
"""

DEFAULT_REQUESTS = {
    'application/json': APPLICATION_JSON_REQUEST,
}

AWS_INTEGRATION = {
    'credentials': "arn:aws:iam::{{ accountId }}:role/{{ apiGateway['role'] }}",
    'httpMethod': 'POST',
    'type': 'aws',
    'uri': "arn:aws:apigateway:{{ region }}:lambda:path/2015-03-31/functions/arn:aws:lambda:{{ region }}:{{ accountId }}:function:{{ Lambda['config']['name'] }}:{{ stage }}/invocations"
}

DOCKER_BUILD_SCRIPT = """\
#!/bin/bash

set -e -x

# cryptography needs libffi-devel and openssl-devel, also install extra build
# dependencies.
yum install -y libffi-devel openssl-devel ${EXTRA_PACKAGES}

PYBIN="/opt/python/${PY_VERSION}/bin"

# Enable compiling against OpenSSL 0.9.8
CRYPTOGRAPHY_ALLOW_OPENSSL_098="true"

${PYBIN}/pip wheel --no-binary :all: -w /wheelhouse -r /src/requirements.txt

# We only want to repair platform wheels that were compiled - not universal
# wheels.
find /wheelhouse -type f \\
    -name "*.whl" \\
    -not -name "*none-any.whl" \\
    -exec auditwheel repair {} -w /wheelhouse/ \;

# Provide SHA1-sum file that we can check against to see if dependencies should
# be rebuilt.
sha1sum /src/requirements.txt | cut -d " " -f 1 > /wheelhouse/sha1sum
"""

DOCKER_INSTALL_SCRIPT = """\
#!/bin/bash

set -e -x

PYBIN="/opt/python/${PY_VERSION}/bin"

${PYBIN}/pip install -t /src/${INSTALL_DIR} \\
    --no-compile \\
    --no-index \\
    --find-links /wheelhouse \\
    -r /src/lambda/requirements.txt
"""
